clear; clc;

%% load data
loaddata;
[Ycore, Yothers] = standardize_data( Ycore, Y4others );

%% MCMC for individual models setup
nsims = 20000;          % retained MCMC draws
burnin = 0.1 * nsims;	% burnin period            
simstep = 20;           % draws to skip for thinning
M = 0500;               % number of IS draws for ML computation
svsims = floor( nsims / simstep );

%% MCMC for composite posterior resampling of non-core parameters
cpnsims = nsims / 2;    % make sure cpnsims / cpsimstep = svsims (otherwise bad things happen)
cpsimstep = simstep / 2;

%% fixed specification parameters
loff = 1;   % VAR lag offset (i.e., first VAR lag is y(t-loff), etc.)
p = 4;      % order of the VAR
nother = 1; % number of variables from Yothers to include in each submodel;
            % note: the current code version must have nother = 1 and
            % will not function properly if this setting is changed

[T, ncore] = size( Ycore );         % dimensions of core variables
n = ncore + nother;                 % number of variables in a submodel
m_all = size( Yothers, 2 );         % total number of submodels
T = T - loff * p;                   % adjust the time dimension for initial values
dates = dates( loff * p + 1:end );  % adjust dates according to the new time dimension
kbeta = n * n * p + n;              % no. of parameter in beta
kgam = n * ( n - 1 ) / 2;           % no. of parameter in gamma_t

%% other algorithm settings
DIC = 1;                % compute BIC and DIC if DIC = 1
ML = 1;                 % compute ML if ML = 1

%% construct data objects
[Y_all, X_all, W_all] = construct_data( Ycore, Yothers, T, n, p, loff, kgam, m_all );
stn = struct( 'idx', [], 'n', ncore, 'p', p );  % to enforce stationary on the VAR in core
                                                % variables only embeded in each sub-model

%% initialize the random number generator
rng( 'shuffle' ); rngstate = rng; % a new state
% load rngstate-001; rng( rngstate ); % a previously saved state

%% initialize parallel processing and the text-based progress monitor
doparallel = setup_parpool();
[txtlen, ppm] = setup_progmons( doparallel, 'Running MCMC: ', m_all );

%% useful quantities for computing conditional MLs and combining draws
coreY = repmat( [true( ncore, 1 ); false( nother, 1 )], T, 1 );
coreX = [repmat( [true; repmat( [true( ncore, 1 ); false( nother, 1 )], p, 1 )], ncore, 1 ); ...
        false( nother * ( 1 + n * p ), 1 )];
otherX = [false( ncore * ( 1 + n * p ), 1 ); true( nother * ( 1 + n * p ), 1 )];
coreW = repmat( [true( ncore, 1 ); false( kgam - ncore, 1 )], T, 1 );
stn.idx = coreX; % to enforce stationarity of embedded VAR in core vars

%% construct priors
[beta0, Vbeta, agam, Vgam, ah, Vh, nugam0, Sgam0, nuh0, Sh0, prior_all, prior_core] = ...
    construct_priors( Y_all, X_all, coreY, otherX, coreW, n, T, p, kbeta, kgam, m_all );

%% allocate space for simulation draws
simresults = struct( 'beta', cell( m_all, 1 ), ...
                     'gam', cell( m_all, 1 ), 'h', cell( m_all, 1 ), ...
                     'Siggam', cell( m_all, 1 ), 'Sigh', cell( m_all, 1 ), ...
                     'gam0', cell( m_all, 1 ), 'h0', cell( m_all, 1 ), ...
                     'mlx', cell( m_all, 1 ), 'mlw', cell( m_all, 1 ), ...
                     'dic', cell( m_all, 1 ) );

%% estimate all sub-models in parallel
disp( ['Starting estimation with ' num2str( m_all ) ' sub-models...'] );
runtime = 0; tic; % start the clock
parfor mi = 1:m_all
    % the following need to be predefined for parfor construction
    mlx = []; mlx_std = [];
    mlw = []; mlw_std = [];
    dic = []; pD = []; Dbar = []; maxllike = [];

    % estimate a sub-model
    y = reshape( Y_all{mi}', [], 1 );
    X = SURform2( X_all{mi}, n );
    W = W_all{mi};

    [s_beta, s_gam, s_h, s_Siggam, s_Sigh, s_gam0, s_h0] ...
        = VARSV_mcmc( y, X, W, T, n, stn, kbeta, kgam, ...
            beta0, Vbeta, agam, Vgam, ah, Vh, nugam0, Sgam0, nuh0, Sh0, ...
            nsims, burnin, svsims, simstep );

    %% compute MLs for evaluating sub-model likelihoods
    [mlx, mlx_std] ...
        = ml_tvp_r1_sv_core( y( coreY ), s_beta( :, ~otherX ), ...
        s_Siggam( :, coreW( 1:kgam ) ), s_Sigh( :, coreY( 1:n ) ), ...
        s_h0( :, coreY( 1:n ) ), s_gam0( :, coreW( 1:kgam ) ), ...
        prior_core, X( coreY, ~otherX ), W( coreY, coreW ), ncore, M );

    %% compute MLs for composite likelihood weights
    if ML == 1
        % marginal likelihood for only the ncore variables in a VAR-SV sub-model
        [mlw, mlw_std] ...
            = ml_tvp_r1_sv_core( y, s_beta, s_Siggam, s_Sigh, s_h0, s_gam0, ...
            prior_all, X, W, ncore, M );
    end

    %% compute BIC and DIC for composite likelihood weights
    if DIC == 1
        % DIC for all (ncore+1) variables in a VAR-SV sub-model
        [dic, pD, Dbar, maxllike] ...
            = dic_tvp_r1_sv_core( y, s_beta, s_Siggam, s_Sigh, s_h0, s_gam0, ...
            X, W, ncore, ceil( svsims / M ) );
    end

    % save sim results for processing after completing all sub-model computation
    simresults( mi ).beta = s_beta;
    simresults( mi ).gam = s_gam;
    simresults( mi ).h = s_h;
    simresults( mi ).Siggam = s_Siggam;
    simresults( mi ).Sigh = s_Sigh;
    simresults( mi ).gam0 = s_gam0;
    simresults( mi ).h0 = s_h0;
    simresults( mi ).mlx = [mlx mlx_std];
    simresults( mi ).mlw = [mlw mlw_std];
    simresults( mi ).dic = [dic pD Dbar maxllike];

    % update the progress monitor
    if ~isempty( ppm )
        ppm.increment();
    end
end

% close the progress window if it exists
if ~isempty( ppm )
    ppm.delete();
end

% compute CL weights
wCL = cell( 4, 1 );
lnML = reshape( [simresults.mlw], 2, m_all );
lnDIC = reshape( [simresults.dic], 4, m_all );

maxlnML = max( lnML( 1, : ) );
maxlnDIC = max( lnDIC( 1, : ) );
maxlnBIC = max( lnDIC( 4, : ) ); % maxllike is the fourth row in lnDIC

wCL{1} = exp( lnML( 1, : )' - maxlnML ) / sum( exp( lnML( 1, : ) - maxlnML ) );
wCL{2} = exp( lnDIC( 1, : )' - maxlnDIC ) / sum( exp( lnDIC( 1, : ) - maxlnDIC ) );
wCL{3} = exp( lnDIC( 4, : )' - maxlnBIC ) / sum( exp( lnDIC( 4, : ) - maxlnBIC ) );
wCL{4} = ones( m_all, 1 ) / m_all;

% pool draws from individual submodels to obtain draws from the
% composite posterior (i.e. derived from the composite likelihood)
s_lnqCL = [];
s_V = cell( size( wCL ) );

for j = 1:length( wCL )
    % resample core parameters using simple rejection sampling
    [b_core, g_core, h_core, sg_core, sh_core, s_lnqCL] ...
                    = VARCL_rs( Y_all, X_all, W_all, beta0, Vbeta, ...
                    simresults, wCL{j}, coreY, coreX, otherX, coreW, ...
                    kgam, T, n, ncore, svsims, doparallel, s_lnqCL );

%         %%%%% TO BE COMPLETED %%%%%
% %         % resample noncore parameters using MCMC
%         kgam_other = nnz( ~coreW( 1:kgam ) );
%         b_other = nan( svsims, nnz( otherX ), m_all );
%         g_other = nan( svsims, nnz( ~coreW ), m_all );
%         h_other = nan( svsims, nother * T, m_all );
%         sg_other = nan( svsims, kgam_other, m_all );
%         sh_other = nan( svsims, nother, m_all );
%         
%         [~, ppm] = setup_progmons( doparallel, 'Running MCMC for noncore params: ', m_all );
%         wCLj = wCL{j};
%         
%         parfor mi = 1:m_all
%             if wCLj( mi ) > 1 / ( m_all * 1e6 )
%                 % estimate a submodel if it has a nontrivial weight
%                 y = reshape( Y_all{mi}', [], 1 );
%                 X = SURform2( X_all{mi}, n );
%                 W = W_all{mi};
% 
%                 % start with draws from sub-model MCMC
%                 startvals = struct( ...
%                     'beta', mean( simresults( mi ).beta( :, otherX ) )', ...
%                     'gam', mean( simresults( mi ).gam( :, ~coreW ) )', ...
%                     'h', reshape( mean( simresults( mi ).h( :, ~coreY ) ), nother, T )', ...
%                     'Siggam', mean( simresults( mi ).Siggam( :, ~coreW( 1:kgam ) ) )', ...
%                     'Sigh', mean( simresults( mi ).Sigh( :, ~coreY( 1:n ) ) )', ...
%                     'gam0', mean( simresults( mi ).gam0( :, ~coreW( 1:kgam ) ) )', ...
%                     'h0', mean( simresults( mi ).h0( :, ~coreY( 1:n ) ) )' );
% 
%                 [b_other( :, :, mi ), g_other( :, :, mi ), h_other( :, :, mi ), ...
%                 sg_other( :, :, mi ), sh_other( :, :, mi )] ...
%                     = VARSV_mcmc( y( ~coreY ), X( ~coreY, otherX ), W( ~coreY, ~coreW ), ...
%                         T, nother, [], nnz( otherX ), nnz( ~coreW( 1:kgam ) ), ...
%                         beta0( otherX ), Vbeta( otherX ), ...
%                         agam( ~coreW( 1:kgam ) ), Vgam( ~coreW( 1:kgam ) ), ...
%                         ah( ~coreY( 1:n ) ), Vh( ~coreY( 1:n ) ), ...
%                         nugam0( ~coreW( 1:kgam ) ), Sgam0( ~coreW( 1:kgam ) ), ...
%                         nuh0( ~coreY( 1:n ) ), Sh0( ~coreY( 1:n ) ), ...
%                         cpnsims, 0, svsims, cpsimstep, startvals, log( wCLj( mi ) ) );
% 
%                 % update the progress monitor
%                 if ~isempty( ppm )
%                     ppm.increment();
%                 end
%             end
%         end
% 
%         % close the progress window if it exists
%         if ~isempty( ppm )
%             ppm.delete();
%         end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%

    % compute covariance matrices
    s_V{j} = recover_V( g_core, h_core, nnz( coreW( 1:kgam ) ), ncore, T, svsims );
end
fprintf( '\nEstimation completed after %5.3f hours.\n', toc / 3600 );